from .iface import Special
from .match import ExactMatch

class Started(Special): pass

def starter(startabletype):
    try:
        starter = startabletype.di_starter
    except AttributeError:
        pass
    else:
        match, = starter.__init__.di_deptypes
        if match.clazz == startabletype: # Otherwise it's inherited.
            return starter
    from .diapyr import types
    @types(ExactMatch(startabletype))
    def __init__(self, startable):
        startable.start()
        self.startable = startable
    def dispose(self):
        self.startable.stop() # FIXME: Untested.
    startabletype.di_starter = startedtype = type("Started[%s]" % Special.gettypelabel(startabletype), (Started,), {f.__name__: f for f in [__init__, dispose]})
    return startedtype



from contextlib import contextmanager
from importlib import import_module
import os, sys

def _patch(modulename, setup):
    try:
        import_module(modulename).setup = setup
    except Exception as e: # XXX: Not BaseException?
        sys.stderr.write("Leave %s unpatched: %s\n" % (modulename, e))

class Stack(list):

    def setup(self, **kwargs):
        self.append(kwargs)

@contextmanager
def _outtoerr():
    stdout, sys.stdout = sys.stdout, sys.stderr
    try:
        yield
    finally:
        sys.stdout = stdout

def main():
    sys.argv.pop(0)
    path = sys.argv[0]
    fields = set(sys.argv[1:])
    sys.argv[1:] = ['--name'] # Invoke as little around setup itself as possible.
    sys.path.insert(0, os.path.dirname(path))
    stack = Stack()
    for m in 'distutils.core', 'setuptools':
        _patch(m, stack.setup)
    try:
        with _outtoerr(), open(path) as f:
            exec(f.read(), dict(__name__ = '__main__', __file__ = path))
    except BaseException as e: # Such as SystemExit for bad interpreter version.
        sys.stdout.write(repr(e))
    else:
        setupkwargs, = stack
        sys.stdout.write(repr({k: v for k, v in setupkwargs.items() if k in fields}))

if ('__main__' == __name__):
    main()



from .diapyr import DI, types
from .iface import MissingAnnotationException, UnsatisfiableRequestException

DI = DI
'Convenience import.'
assert MissingAnnotationException
types = types
'Convenience import.'
assert UnsatisfiableRequestException



from .iface import ImpasseException, MissingAnnotationException, unset
from .match import AllInstancesOf, wrap
from .source import Builder, Class, Factory, Instance, Proxy
from .start import starter
from .util import bfs, invokeall, singleton
from collections import defaultdict, OrderedDict
from inspect import isfunction
import logging

log = logging.getLogger(__name__)
typeitself = type

@singleton
class NullPlan:

    args = ()

    def make(self):
        pass

def types(*deptypes, **kwargs):
    'Declare that the decorated function or method expects args of the given types. Use square brackets to request all instances of a type as a list. Use `this` kwarg to declare the type of result returned by a factory.'
    def g(f):
        f.di_deptypes = tuple(wrap(t) for t in deptypes)
        if 'this' in kwargs:
            f.di_owntype = kwargs['this']
        return f
    return g

class DI:

    log = log # Tests may override.
    depthunit = '>'

    def __init__(self, parent = None):
        self.typetosources = defaultdict(list)
        self.allsources = [] # Old-style classes won't be registered against object.
        self.parent = parent

    def addsource(self, source):
        for type in source.types:
            self.typetosources[type].append(source)
        self.allsources.append(source)

    def removesource(self, source): # TODO: Untested.
        for type in source.types:
            self.typetosources[type].remove(source)
        self.allsources.remove(source)

    def addclass(self, clazz):
        try:
            clazz.__init__.di_deptypes
        except AttributeError:
            raise MissingAnnotationException("Missing types annotation: %s" % clazz)
        self.addsource(Class(clazz, self))
        self._addbuilders(clazz)
        if getattr(clazz, 'start', None) is not None:
            self.addclass(starter(clazz))

    def _addbuilders(self, cls):
        for name in dir(cls):
            m = getattr(cls, name)
            if hasattr(m, 'di_deptypes') and hasattr(m, 'di_owntype'):
                assert '__init__' != name # TODO LATER: Check upfront.
                self.addsource(Builder(cls, m, self))

    def addinstance(self, instance, type = None):
        clazz = instance.__class__ if type is None else type
        self.addsource(Instance(instance, clazz))
        if not isinstance(instance, typeitself):
            self._addbuilders(clazz)

    def addfactory(self, factory):
        self.addsource(Factory(factory, self))

    def _addmethods(self, obj):
        if hasattr(obj, 'di_owntype') or isfunction(obj):
            yield self.addfactory
        elif hasattr(obj, '__class__'):
            clazz = obj.__class__
            if clazz == type: # It's a non-fancy class.
                yield self.addclass
            elif isinstance(obj, type): # It's a fancy class.
                yield self.addinstance
                try:
                    obj.__init__.di_deptypes
                    yield self.addclass
                except AttributeError:
                    pass # Not admissible as class.
            else: # It's an instance.
                yield self.addinstance
        else: # It's an old-style class.
            yield self.addclass

    def add(self, obj):
        'Register the given class, factory or instance.'
        for m in self._addmethods(obj):
            m(obj)

    def all(self, type):
        'Return all objects of the given type, instantiating them and collaborators if necessary.'
        return self._session(AllInstancesOf(type))

    def __call__(self, clazz):
        'Return unique object of the given type, instantiating it and its collaborators if necessary.'
        return self._session(wrap(clazz))

    def _session(self, match):
        root = match.di_get(self, unset)
        plans = OrderedDict()
        @bfs([root])
        def proc(info, a):
            for s in a.sources:
                if s not in plans:
                    p = s.plan(self.depthunit * info.depth, a.trigger)
                    if p is None:
                        p = NullPlan
                    plans[s] = p
                    for b in p.args:
                        yield b
        while plans:
            sources = [s for s in plans if not any(r in plans for a in plans[s].args for r in a.sources)]
            if not sources:
                raise ImpasseException
            for s in sources:
                plans.pop(s).make()
        return root.resolve()

    def join(self, type, discardall = True):
        self.parent.addsource(Proxy(self, type, discardall))

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        'Discard all instances created by this container, calling `dispose` if they have it.'
        self.discardall()

    def discardall(self):
        invokeall([s.discard for s in reversed(self.allsources)])



from inspect import isfunction
import sys

ispy2 = sys.version_info.major < 3

class Proxy(object):

    def __getattr__(self, name):
        try:
            return getattr(self._enclosinginstance, name)
        except AttributeError:
            superclass = super(Proxy, self)
            try:
                supergetattr = superclass.__getattr__
            except AttributeError:
                raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
            return supergetattr(name)

def innerclass(cls):
    'An instance of the decorated class may access its enclosing instance via `self`.'
    class InnerMeta(type):
        def __get__(self, enclosinginstance, owner):
            clsname = (cls if self is Inner else self).__name__
            return type(clsname, (Proxy, self), dict(_enclosinginstance = enclosinginstance))
    Inner = InnerMeta('Inner', (cls,), {})
    return Inner

def singleton(t):
    'The decorated class is replaced with a no-arg instance.'
    return t()

@singleton
class outerzip:

    class Session:

        def __init__(self, iterables):
            self.iterators = [iter(i) for i in iterables]

        def row(self):
            self.validrow = len(self.iterators)
            for i in self.iterators:
                try:
                    yield next(i)
                except StopIteration:
                    self.validrow -= 1
                    yield

    def __call__(self, *iterables):
        session = self.Session(iterables)
        while True:
            values = tuple(session.row())
            if not session.validrow:
                break
            yield values

def enum(*lists):
    def d(cls):
        cls.enum = v = []
        for args in lists:
            obj = cls(*args)
            setattr(cls, args[0], obj)
            v.append(obj)
        return cls
    return d

def _rootcontext(e):
    while True:
        c = getattr(e, '__context__', None)
        if c is None:
            return e
        e = c

def invokeall(callables):
    '''Invoke every callable, even if one or more of them fail. This is mostly useful for synchronising with futures.
    If all succeeded return their return values as a list, otherwise raise all exceptions thrown as a chain.'''
    values = []
    failure = None
    for c in callables:
        try:
            obj = c()
        except Exception as e:
            _rootcontext(e).__context__ = failure
            failure = e
        else:
            values.append(obj)
    if failure is None:
        return values
    raise failure

def bfs(keys):
    '''Breadth-first search starting with the given iterable of keys, intended to be used as a decorator.
    If a function is decorated it should take an info object and key, and yield subsequent keys.
    If a class is decorated, a new instance of it is used as info object:
    The class should have a `newdepth` method that will be called before each depth, and a `process` method that takes a key and yields subsequent keys as in the function case.
    The info object is kept updated with the list of `currentkeys`, current `depth` and the set of `donekeys`.
    Note that the first `currentkeys` (`depth` 0) is exactly the passed in `keys` iterable, subsequent `currentkeys` will be non-empty lists.
    The process function is only invoked for keys that have not yet been processed, i.e. unique keys.
    When finished the decorated function/class is replaced with the last state of the info object.'''
    def transform(function_or_class):
        if isfunction(function_or_class):
            class Info:
                def newdepth(self):
                    pass
                def process(self, key):
                    return function_or_class(self, key)
            info = Info()
        else:
            info = function_or_class()
        nextkeys = keys
        info.depth = -1
        info.donekeys = set()
        while True:
            iterator = iter(nextkeys)
            try:
                key = next(iterator)
            except StopIteration:
                break
            info.currentkeys = nextkeys
            info.depth += 1
            info.newdepth()
            nextkeys = []
            while True:
                if key not in info.donekeys:
                    j = info.process(key)
                    if j is not None:
                        nextkeys.extend(j)
                    info.donekeys.add(key)
                try:
                    key = next(iterator)
                except StopIteration:
                    break
        return info
    return transform
